from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any

import iso639
import pycountry
import pycountry.db

from rapid_tasty_api.tag import Tag

from .credit import Credit
from .instruction import Instruction
from .nutrition import Nutrition
from .price import Price
from .ratings import Ratings
from .section import Section
from .topic import Topic


@dataclass(frozen=True, slots=True)
class RecipeMetadata:
    """Represents the metadata of a recipe."""

    approved_at: datetime
    aspect_ratio: str
    beauty_url: str | None
    brand: str | None
    brand_id: int | None
    buzz_id: int | None
    canonical_id: str
    # TODO (when I find an example): compilations: list[SOMETHING]
    country: pycountry.db.Country
    created_at: datetime
    credits: list[Credit]
    draft_status: str
    # TODO (when I find an example): facebook_posts: list[SOMETHING]
    id: int
    inspired_by_url: str | None
    is_app_only: bool
    is_one_top: bool
    is_shoppable: bool
    is_subscriber_content: bool
    keywords: str
    language: iso639.Language
    nutrition_visibility: str
    original_video_url: str | None
    promotion: str
    # TODO (when I find an example): renditions: list[SOMETHING]
    seo_path: str
    seo_title: str
    servings_noun_plural: str
    servings_noun_singular: str
    show_name: str
    show_id: int
    slug: str
    thumbnail_alt_text: str
    thumbnail_url: str
    tips_and_ratings_enabled: bool
    # TODO (when I find an example): total_time_tier: SOMETHING | None
    updated_at: datetime
    # TODO (when I find an example): video_ad_content: SOMETHING | None
    video_id: int | None
    video_url: str | None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> RecipeMetadata:
        return RecipeMetadata(
            approved_at=datetime.fromtimestamp(data["approved_at"]),
            aspect_ratio=data["aspect_ratio"],
            beauty_url=data["beauty_url"],
            brand=data["brand"],
            brand_id=data["brand_id"],
            buzz_id=data["buzz_id"],
            canonical_id=data["canonical_id"],
            country=pycountry.countries.get(alpha_2=data["country"]),  # type: ignore
            created_at=datetime.fromtimestamp(data["created_at"]),
            credits=[Credit.from_dict(credit) for credit in data["credits"]],
            draft_status=data["draft_status"],
            id=data["id"],
            inspired_by_url=data["inspired_by_url"],
            is_app_only=data["is_app_only"],
            is_one_top=data["is_one_top"],
            is_shoppable=data["is_shoppable"],
            is_subscriber_content=data["is_subscriber_content"],
            keywords=data["keywords"],
            language=iso639.Language.from_part3(data["language"]),
            nutrition_visibility=data["nutrition_visibility"],
            original_video_url=data["original_video_url"],
            promotion=data["promotion"],
            seo_path=data["seo_path"],
            seo_title=data["seo_title"],
            servings_noun_plural=data["servings_noun_plural"],
            servings_noun_singular=data["servings_noun_singular"],
            show_name=data["show"]["name"],
            show_id=data["show_id"],
            slug=data["slug"],
            thumbnail_alt_text=data["thumbnail_alt_text"],
            thumbnail_url=data["thumbnail_url"],
            tips_and_ratings_enabled=data["tips_and_ratings_enabled"],
            updated_at=datetime.fromtimestamp(data["updated_at"]),
            video_id=data["video_id"],
            video_url=data["video_url"],
        )


@dataclass(frozen=True, slots=True)
class Recipe:
    """Represents a recipe."""

    metadata: RecipeMetadata
    cook_time_minutes: int | None
    description: str
    instructions: list[Instruction]
    name: str
    num_servings: int
    nutrition: Nutrition
    prep_time_minutes: int | None
    price: Price
    sections: list[Section]
    tags: list[Tag]
    topics: list[Topic]
    total_time_minutes: int | None
    user_ratings: Ratings
    yields: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Recipe:
        cook_time_minutes: int | None = data.pop("cook_time_minutes")
        desciption: str = data.pop("description")
        instructions: list[Instruction] = [
            Instruction.from_dict(instruction)
            for instruction in data.pop("instructions")
        ]
        name: str = data.pop("name")
        num_servings: int = data.pop("num_servings")
        nutrition: Nutrition = Nutrition.from_dict(data.pop("nutrition"))
        prep_time_minutes: int | None = data.pop("prep_time_minutes")
        price: Price = Price.from_dict(data.pop("price"))
        sections: list[Section] = [
            Section.from_dict(section) for section in data.pop("sections")
        ]
        tags: list[Tag] = [Tag.from_dict(tag) for tag in data.pop("tags")]
        topics: list[Topic] = [Topic.from_dict(topic) for topic in data.pop("topics")]
        total_time_minutes: int | None = data.pop("total_time_minutes")
        user_ratings: Ratings = Ratings.from_dict(data.pop("user_ratings"))
        yields: str = data.pop("yields")
        metadata: RecipeMetadata = RecipeMetadata.from_dict(data)

        return Recipe(
            metadata,
            cook_time_minutes,
            desciption,
            instructions,
            name,
            num_servings,
            nutrition,
            prep_time_minutes,
            price,
            sections,
            tags,
            topics,
            total_time_minutes,
            user_ratings,
            yields,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from .ingredient import Ingredient
from .measurement import Measurement


class IncompatibleComponentError(ValueError):
    pass


@dataclass(frozen=True, slots=True)
class Component:
    """
    Collection of details surrounding an ingredient.
    Ex: 10 cloves peeled garlic
    """

    extra_comment: str
    id: int
    ingredient: Ingredient
    measurements: list[Measurement]
    position: int
    raw_text: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Component:
        return Component(
            extra_comment=data["extra_comment"],
            id=data["id"],
            ingredient=Ingredient.from_dict(data["ingredient"]),
            measurements=[
                Measurement.from_dict(measurement)
                for measurement in data["measurements"]
            ],
            position=data["position"],
            raw_text=data["raw_text"],
        )

    def __add__(self, other: Component) -> Component:
        if self.ingredient.id != other.ingredient.id:
            raise IncompatibleComponentError(
                "Components must be the same in order to add their amounts."
            )

        ret = Component(
            extra_comment="",
            id=self.id,
            ingredient=self.ingredient,
            measurements=[],
            position=0,
            raw_text="",
        )

        for measurement in self.measurements:
            other_measurement_quantity = 0
            for other_measurement in other.measurements:
                if measurement.unit.name != other_measurement.unit.name:
                    continue
                other_measurement_quantity = other_measurement.quantity
                break

            ret.measurements.append(
                Measurement(
                    id=measurement.id,
                    quantity=measurement.quantity + other_measurement_quantity,
                    unit=measurement.unit,
                )
            )

        return ret



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from unicodedata import numeric


def parse_number(number: str) -> float:
    if number.isascii() and number.isnumeric():
        return float(number)
    elif len(number) > 1:
        # stole this code from stackoverflow: https://stackoverflow.com/a/50264056
        return float(number[:-1]) + numeric(number[-1])
    return numeric(number)


@dataclass(frozen=True, slots=True)
class Unit:
    """Represents the unit of a `Measurement`."""

    abbreviation: str
    display_plural: str
    display_singular: str
    name: str
    system: str | None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Unit:
        return Unit(
            abbreviation=data["abbreviation"],
            display_plural=data["display_plural"],
            display_singular=data["display_singular"],
            name=data["name"],
            system=data.get("system"),
        )


@dataclass(frozen=True, slots=True)
class Measurement:
    """Represents the amount of a `Component`."""

    id: int
    quantity: float
    unit: Unit

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Measurement:
        return Measurement(
            id=data["id"],
            quantity=parse_number(data["quantity"]),
            unit=Unit.from_dict(data["unit"]),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True, slots=True)
class Credit:
    """Represents a credit (eg. author) of the recipe."""

    name: str | None
    type: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Credit:
        return Credit(
            name=data["name"],
            type=data["type"],
        )



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass(frozen=True, slots=True)
class Price:
    """Represents the price of a recipe in cents (probably USD)."""

    consumption_portion: int
    consumption_total: int
    portion: int
    total: int
    updated_at: datetime

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Price:
        return Price(
            consumption_portion=data["consumption_portion"],
            consumption_total=data["consumption_total"],
            portion=data["portion"],
            total=data["total"],
            updated_at=datetime.fromisoformat(data["updated_at"]),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True, slots=True)
class Instruction:
    """Represents a single step in the recipe's instructions."""

    appliance: str | None
    display_text: str
    end_time: int
    id: int
    position: int
    start_time: int
    # TODO (when I find an example): temperature: SOMETHING | None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Instruction:
        return Instruction(
            appliance=data["appliance"],
            display_text=data["display_text"],
            end_time=data["end_time"],
            id=data["id"],
            position=data["position"],
            start_time=data["start_time"],
        )



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class Completion:
    """Represents the result of an auto complete search."""

    value: str
    type: str

    @classmethod
    def from_dict(cls, data: dict[str, str]) -> Completion:
        return Completion(
            value=data["display"],
            type=data["type"],
        )



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class Topic:
    """Represents a topics of the recipe."""

    name: str
    slug: str

    @classmethod
    def from_dict(cls, data: dict[str, str]) -> Topic:
        return Topic(name=data["name"], slug=data["slug"])



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass(frozen=True, slots=True)
class TipMetadata:
    """Represents the metadata of a Tip object."""

    author_avatar_url: str
    author_rating: int
    author_user_id: int
    author_is_verified: int
    is_flagged: bool
    recipe_id: int
    status_id: int
    comment_id: int
    comment_count: int
    tip_id: int
    created_at: datetime | None
    updated_at: datetime

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> TipMetadata:
        return TipMetadata(
            author_avatar_url=data["author_avatar_url"],
            author_rating=data["author_rating"],
            author_user_id=data["author_user_id"],
            author_is_verified=data["author_is_verified"],
            is_flagged=data["is_flagged"],
            recipe_id=data["recipe_id"],
            status_id=data["status_id"],
            comment_id=data["comment_id"],
            comment_count=data["comment_count"],
            tip_id=data["tip_id"],
            created_at=(
                datetime.fromtimestamp(data["created_at"])
                if data["created_at"] is not None
                else None
            ),
            updated_at=datetime.fromtimestamp(data["updated_at"]),
        )


@dataclass(frozen=True, slots=True)
class Tip:
    """Represents a tip (review) for a specific recipe."""

    metadata: TipMetadata
    author_name: str
    author_username: str
    tip_body: str
    upvotes_total: int

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Tip:
        author_name: str = data.pop("author_name")
        author_username: str = data.pop("author_username")
        tip_body: str = data.pop("tip_body")
        upvotes_total: int = data.pop("upvotes_total")
        metadata: TipMetadata = TipMetadata.from_dict(data)

        return Tip(metadata, author_name, author_username, tip_body, upvotes_total)



from datetime import UTC, time, tzinfo
from enum import StrEnum

import requests

from rapid_tasty_api._constants import TASTY_HOST
from rapid_tasty_api.data import (
    CompletionData,
    RecipeListData,
    TagListData,
    TipListData,
)
from rapid_tasty_api.feed import Feed
from rapid_tasty_api.recipe import Completion, Recipe
from rapid_tasty_api.tag import Tag, tag_list_to_str
from rapid_tasty_api.tip import Tip


class SortingMethod(StrEnum):
    POPULAR = "popular"
    DESCENDING = "approved_at:desc"
    ASCENDING = "approved_at:asc"


def _timezone_to_utc_offset(timezone_: tzinfo) -> str:
    return time(tzinfo=timezone_).strftime("%z")


class Client:
    """A client to send requests to the Tasty API."""

    __slots__ = "_session"

    def __init__(self, rapid_api_key: str) -> None:
        """
        A client to send requests to the Tasty API.

        :param str rapid_api_key: Your API key. Get one from https://rapidapi.com/apidojo/api/tasty.
        """

        self._session = requests.Session()
        self._session.headers.update(
            {
                "X-RapidAPI-Key": rapid_api_key,
                "X-RapidAPI-Host": TASTY_HOST,
            }
        )

    def get_recipes_auto_complete(self, prefix: str) -> list[Completion]:
        """
        Get auto complete suggestions by name or ingredients.

        :param str prefix: The text to be auto completed.
        :return list[Completion]: A list of possible auto completions.
        """

        url = "https://tasty.p.rapidapi.com/recipes/auto-complete"

        querystring = {"prefix": prefix}

        response = self._session.get(url, params=querystring)

        data = CompletionData.from_dict(response.json())

        return data.results

    def get_recipes_list(
        self,
        offset: int,
        size: int,
        tags: list[Tag] | None = None,
        query: str | None = None,
        sort: SortingMethod = SortingMethod.POPULAR,
    ) -> list[Recipe]:
        """
        Get a list of recipes.

        :param int offset: The amount of recipes to skip.
        :param int size: The amount of recipes to get.
        :param list[Tag] | None tags: A list of tags you want to search for, defaults to None
        :param str | None query: Name of food or ingredients to search by, defaults to None
        :param SortingMethod sort: The method of sorting the results, defaults to SortingMethod.POPULAR
        :return list[Recipe]: The list of recipes.
        """

        url = "https://tasty.p.rapidapi.com/recipes/list"

        querystring = {"from": str(offset), "size": str(size)}

        if tags is not None:
            querystring.update({"tags": tag_list_to_str(tags)})

        if query is not None:
            querystring.update({"q": query})

        if sort != SortingMethod.POPULAR:
            querystring.update({"sort": sort})

        response = self._session.get(url, params=querystring)

        data = RecipeListData.from_dict(response.json())

        return data.results

    def get_recipes_list_similarities(self, recipe_id: int) -> list[Recipe]:
        """
        Get a list of recipes similar to a given recipe.

        :param int recipe_id: The recipe to get similar recipes to.
        :return list[Recipe]: The list of similar recipes.
        """

        url = "https://tasty.p.rapidapi.com/recipes/list-similarities"

        querystring = {"recipe_id": str(recipe_id)}

        response = self._session.get(url, params=querystring)

        data = RecipeListData.from_dict(response.json())

        return data.results

    def get_recipes_more_info(self, recipe_id: int) -> Recipe:
        """
        Get a specific recipe.

        :param int recipe_id: The recipe to get information on.
        :return Recipe: The recipe.
        """

        url = "https://tasty.p.rapidapi.com/recipes/get-more-info"

        querystring = {"recipe_id": recipe_id}

        response = self._session.get(url, params=querystring)

        recipe = Recipe.from_dict(response.json())

        return recipe

    def get_feeds_list(
        self,
        vegetarian: bool,
        timezone: tzinfo = UTC,
        offset: int = 0,
        size: int = 5,
    ) -> list[Feed]:
        """
        Get a list of the latests feeds.

        Feeds are lists of specifically categorized recipes.

        :param int offset: The amount of feeds to skip.
        :param int size: The amount of feeds to get.
        :param bool vegetarian: List vegetarian recipes only.
        :param tzinfo timezone: Your timezone. Defaults to UTC.
        :return list[Feed]: The list of feeds.
        """

        url = "https://tasty.p.rapidapi.com/feeds/list"

        querystring = {
            "size": str(size),
            "timezone": _timezone_to_utc_offset(timezone),
            "vegetarian": "true" if vegetarian else "false",
            "from": str(offset),
        }

        response = self._session.get(url, params=querystring)

        data = [Feed.from_dict(result) for result in response.json()["results"]]

        return data

    def get_tags_list(self) -> list[Tag]:
        """
        Get a list of all tags.

        :return list[Tag]: The list of tags.
        """

        url = "https://tasty.p.rapidapi.com/tags/list"

        response = self._session.get(url)

        data = TagListData.from_dict(response.json())

        return data.results

    def get_tips_list(
        self, recipe_id: int, offset: int = 0, size: int = 30
    ) -> list[Tip]:
        """
        Get a list of tips (reviews) pertaining to a recipe.

        :param int recipe_id: The recipe ID.
        :param int offset: The amount of tips to skip.
        :param int size: The amount of tips to get.

        :return list[Tip]: The list of tips.
        """

        url = "https://tasty.p.rapidapi.com/tips/list"

        querystring = {
            "id": str(recipe_id),
            "from": str(offset),
            "size": str(size),
        }

        response = self._session.get(url, params=querystring)

        data = TipListData.from_dict(response.json())

        return data.results



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True, slots=True)
class Ratings:
    count_negative: int
    count_positive: int
    score: float

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Ratings:
        return Ratings(
            count_negative=data["count_negative"],
            count_positive=data["count_positive"],
            score=data["score"],
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from rapid_tasty_api.recipe import Recipe


@dataclass(frozen=True, slots=True)
class Feed:
    """Represents one of Tasty's feeds (e.g. Popular Recipes This Week)"""

    type: str
    items: list[Recipe]
    name: str | None = None
    category: str | None = None
    min_items: int | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Feed:
        if data["type"] in ("item", "featured"):
            return Feed(
                type=data["type"],
                items=[Recipe.from_dict(data["item"])],
            )

        # Uses .get() because sometimes the values don't exist.

        return Feed(
            type=data["type"],
            items=[Recipe.from_dict(item) for item in data["items"]],
            name=data.get("name"),
            category=data.get("category"),
            min_items=data.get("min_items"),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from rapid_tasty_api.recipe import Completion, Recipe
from rapid_tasty_api.tag import Tag
from rapid_tasty_api.tip import Tip


@dataclass(frozen=True, slots=True)
class CompletionData:
    """Custom class to destructure auto complete requests."""

    results: list[Completion]

    @classmethod
    def from_dict(cls, data: dict[str, list[dict[str, str]]]) -> CompletionData:
        return CompletionData(
            [Completion.from_dict(result) for result in data["results"]]
        )


@dataclass(frozen=True, slots=True)
class RecipeListData:
    """Custom class to destructure recipes/list requests."""

    count: int
    results: list[Recipe]

    # TODO (maybe): Replace the Any with the full type of the dictionary. Big maybe.
    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> RecipeListData:
        return RecipeListData(
            data["count"], [Recipe.from_dict(result) for result in data["results"]]
        )


@dataclass(frozen=True, slots=True)
class TagListData:
    """Custom class to destructure recipes/list requests."""

    count: int
    results: list[Tag]

    # Can someone figure out how to use these? The endpoints don't seem to exist.
    prev: str | None
    next: str | None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> TagListData:
        return TagListData(
            data["count"],
            [Tag.from_dict(result) for result in data["results"]],
            data["prev"],
            data["next"],
        )


@dataclass(frozen=True, slots=True)
class TipListData:
    """Custom class to destructure recipes/list requests."""

    count: int
    results: list[Tip]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> TipListData:
        return TipListData(
            data["count"], [Tip.from_dict(result) for result in data["results"]]
        )



from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
from typing import Any


@dataclass(frozen=True, slots=True)
class Tag:
    """Represents a tag of a recipe."""

    id: int
    type: str
    name: str
    display_name: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Tag:
        return Tag(
            id=data["id"],
            type=data["type"],
            name=data["name"],
            display_name=data["display_name"],
        )


def tag_list_to_str(tags: list[Tag]) -> str:
    """
    Helper function to turn a list of Tag objects into a TastyAPI compatible string. Used when filtering requests by tags.

    :param list[Tag] tags: The list of Tag objects.
    :return str: A formatted string containing all the tags, ready for use in an requests.
    """

    buffer = StringIO()

    for i, tag in enumerate(tags):
        buffer.write(tag.name)
        if i != len(tags) - 1:
            buffer.write(",")

    ret = buffer.getvalue()
    buffer.close()

    return ret



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass(frozen=True, slots=True)
class Ingredient:
    """Represents an ingredient."""

    created_at: datetime
    display_plural: str
    display_singular: str
    id: int
    name: str
    updated_at: datetime

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Ingredient:
        return Ingredient(
            created_at=datetime.fromtimestamp(data["created_at"]),
            display_plural=data["display_plural"],
            display_singular=data["display_singular"],
            id=data["id"],
            name=data["name"],
            updated_at=datetime.fromtimestamp(data["updated_at"]),
        )



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass(frozen=True, slots=True)
class Nutrition:
    """Represents the nutrition facts of a recipe. All values are in grams."""

    calories: int | None
    carbohydrates: int | None
    fat: int | None
    fiber: int | None
    protein: int | None
    sugar: int | None
    updated_at: datetime | None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Nutrition:
        return Nutrition(
            calories=data.get("calories"),
            carbohydrates=data.get("carbohydrates"),
            fat=data.get("fat"),
            fiber=data.get("fiber"),
            protein=data.get("protein"),
            sugar=data.get("sugar"),
            updated_at=datetime.fromisoformat(data["updated_at"])
            if data.get("updated_at") is not None
            else None,
        )



"""`Recipe` and supporting classes"""

from .completion import Completion
from .component import Component
from .credit import Credit
from .ingredient import Ingredient
from .instruction import Instruction
from .measurement import Measurement, Unit
from .nutrition import Nutrition
from .price import Price
from .ratings import Ratings
from .recipe import Recipe, RecipeMetadata
from .section import Section
from .topic import Topic

__all__ = [
    "Completion",
    "Component",
    "Credit",
    "Ingredient",
    "Instruction",
    "Measurement",
    "Nutrition",
    "Price",
    "Ratings",
    "Recipe",
    "RecipeMetadata",
    "Section",
    "Topic",
    "Unit",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from .component import Component


@dataclass(frozen=True, slots=True)
class Section:
    """Represents a part of a recipe."""

    components: list[Component]
    name: str | None
    position: int

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Section:
        return Section(
            components=[
                Component.from_dict(component) for component in data["components"]
            ],
            name=data["name"],
            position=data["position"],
        )
